"""
Plugin with goals to build dependencies, tools, sources, and resources.
"""
import logging
import shutil
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Tuple

from caiman.config import Command, Config
from caiman.installer import DependencyInstaller, ToolInstaller
from caiman.plugins.base import Goal, Plugin, param
from caiman.source import WorkspacePythonSource, WorkspaceSource


@dataclass
class BuildCommand:
    """
    Defines the schema for the build command.
    """
    target: str = param("Target to build", default="")
    force: bool = param("Force build", default=False)

    @property
    def builder(self):
        """
        Return the builder name from the target
        """
        return self.target.split(":", 1)[0] if self.target else None

    @property
    def buildable(self):
        """
        Return the buildable name from the target
        """
        parts = self.target.split(":", 1)
        return self.target.split(":", 1)[1] if len(parts) > 1 else None


class Builder(ABC):
    """
    Base class for building sources.
    """
    def __init__(self, config: Config):
        """
        Initialize the builder with the given project configuration.
        :param config: The project configuration
        """
        self.config = config
        self._logger = logging.getLogger(f"build:{self.name}")

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def buildables(self):
        yield from []

    def get_command_buildables(self, command: BuildCommand):
        """
        Return the buildables for the given command by filtering all the
        buildables by the target and buildable name.
        """
        if command.buildable:
            return [
                buildable
                for buildable in self.buildables
                if command.buildable == buildable.source.name
            ]
        return list(self.buildables)

    @abstractmethod
    def _build(self, source: WorkspaceSource, command: BuildCommand):
        """
        Process a buildable source.
        """

    def __call__(self, command: BuildCommand):
        """
        Execute the build command.
        :param command: The build command derived from the command line arguments
        """
        buildables = self.get_command_buildables(command)
        if not buildables:
            if not command.builder:
                return
            raise RuntimeError(
                f"No buildable sources found for target '{command.target}'"
            )

        self._logger.info(
            f"Building targets: {', '.join(buildable.source.name for buildable in buildables)}"
        )
        for buildable in buildables:
            if not command.buildable or command.buildable == buildable.source.name:
                self._build(buildable, command=command)


class ResourceBuilder(Builder):
    """
    Builder for building resources.
    """
    @property
    def name(self):
        return "resources"

    @property
    def buildables(self):
        yield from [
            WorkspaceSource(workspace=self.config.workspace, source=source)
            for source in self.config.resources
        ]

    def _build(self, source: WorkspaceSource, command: BuildCommand):
        """
        Build the given resource.
        """
        self._logger.info(f"Building {source.source.name}")
        source.manifests.save(source.create_manifest())
        deployment = source.create_deployment()
        manifest = deployment(logger=self._logger)
        self._logger.info(f"Saving {source.source.name} manifest")
        deployment.manifests.save(manifest)


class SourceBuilder(ResourceBuilder):
    """
    Builder for building Python sources.
    """
    @property
    def name(self):
        return "sources"

    @property
    def buildables(self):
        yield from [
            WorkspacePythonSource(workspace=self.config.workspace, source=source)
            for source in self.config.sources
        ]


class DependencyBuilder(SourceBuilder):
    """
    Builder for installing dependencies.
    """
    @property
    def name(self):
        return "dependencies"

    @property
    def buildables(self):
        yield from [
            DependencyInstaller(config=self.config, dependency=dependency)
            for dependency in self.config.dependencies
        ]

    def _build(self, dependency: DependencyInstaller, command: BuildCommand):
        return dependency(force=command.force, logger=self._logger)


class ToolBuilder(DependencyBuilder):
    """
    Builder for installing tools.
    """
    @property
    def name(self):
        return "tools"

    @property
    def buildables(self):
        yield from [
            ToolInstaller(config=self.config, dependency=dependency)
            for dependency in self.config.tools
        ]


class BuildGoal(Goal):
    """
    Goal for building dependencies, tools, sources, and resources.
    """
    @property
    def help(self):
        return "Build dependencies, tools, sources, and resources"

    @property
    def name(self):
        return "build"

    def get_schema(self):
        return BuildCommand

    @property
    def builders(self):
        return (
            ResourceBuilder(self.config),
            SourceBuilder(self.config),
            DependencyBuilder(self.config),
            ToolBuilder(self.config),
        )

    def clean(self):
        mp_deploy_path = self.config.workspace.get_build_asset_path(is_frozen=False)
        if mp_deploy_path.exists():
            self.info(f"Removing {mp_deploy_path}")
            shutil.rmtree(mp_deploy_path, ignore_errors=True)

        frozen_path = self.config.workspace.get_build_asset_path(is_frozen=True)
        if frozen_path.exists():
            self.info(f"Removing {frozen_path}")
            shutil.rmtree(frozen_path, ignore_errors=True)

    def __call__(self, command: Command):
        goal_command = BuildCommand(**command.params)
        if not goal_command.target:
            self.clean()

        for builder in self.builders:
            if not goal_command.builder or goal_command.builder == builder.name:
                try:
                    builder(goal_command)
                except Exception as e:
                    self._logger.exception(f"Error building {builder.name}: {e}")
                    self.fail(str(e))


class ApplicationBuilderPlugin(Plugin):
    """
    Plugin for building dependencies, tools, sources, and resources.
    """
    def get_goals(self) -> Tuple[Goal]:
        return (BuildGoal(self.config),)



"""
Wrappers for package managers to install dependencies and tools
"""
import logging
from dataclasses import dataclass
from pathlib import Path

from caiman.config import Config, Dependency
from caiman.manifest import (
    DependencyManifestRegistry,
    Manifest,
    ManifestItem,
    ToolManifestRegistry,
)
from caiman.proc.base import MicroPythonProcess
from caiman.proc.local import LocalMicroPythonProcess
from caiman.source import (
    WorkspaceDependencySource,
    WorkspaceSource,
    WorkspaceToolSource,
)
from caiman.task import MIPTask, MoveTask

_logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True)
class DependencyInstaller:
    """
    Models a dependency installation process
    """
    config: Config
    dependency: Dependency

    @property
    def workspace(self):
        return self.config.workspace

    @property
    def artifact_root(self):
        """
        Dependency files initially installed to a temporary artifact directory
        that is later scanned for building a manifest of the installed files
        """
        return (
            self.workspace.get_artifact_path("dependencies")
            / self.dependency.package_name
        )

    @property
    def install_root(self):
        """
        Local package installation directory where installed artifacts are moved
        after a manifest is created
        """
        return self.workspace.get_package_path()

    @property
    def manifests(self):
        """
        The manifest registry for the installed dependencies
        """
        return DependencyManifestRegistry(workspace=self.workspace, asset_type="source")

    @property
    def source(self) -> WorkspaceSource:
        """
        Models the sources of the installed dependency
        """
        return WorkspaceDependencySource(
            workspace=self.workspace, source=self.dependency
        )

    def create_manifest(self):
        """
        Scans the artifact directory where the dependency was installed to create a manifest
        that will be used to track the installed files and copy them to the local package directory
        """
        files = [
            p.relative_to(self.artifact_root)
            for p in Path(self.artifact_root).rglob("**/*")
            if p.is_file()
        ]
        items = ManifestItem.from_paths(files, self.artifact_root)
        return Manifest(
            name=self.dependency.package_name,
            version=self.dependency.version,
            items=items,
        )

    def get_tasks(self):
        """
        Get the tasks for installing the dependency.
        These will include a package installation task for each asset in the dependency
        """
        channel = self.config.get_channel(self.dependency.channel)

        install_names = (
            [self.dependency.name]
            if not self.dependency.files
            else [f"{self.dependency.name}/{f}" for f in self.dependency.files]
        )
        install_targets = (
            [Path("")]
            if not self.dependency.files
            else [Path(f).parent for f in self.dependency.files]
        )
        packages_by_target = {}
        for name, target in zip(install_names, install_targets):
            packages_by_target.setdefault(target, {})[name] = self.dependency.version

        for target, packages in packages_by_target.items():
            yield from MIPTask.from_package_dict(
                workspace=self.workspace,
                packages=packages,
                index=channel.index,
                root=self.artifact_root,
                target=target,
            )

    @property
    def handler(self) -> MicroPythonProcess:
        """
        The handler for the MicroPython package manager used to install the dependency
        """
        return LocalMicroPythonProcess()

    def install(self, force=False, logger=None):
        """
        Install the dependency.
        A manifest is created for the installed files and saved to the manifest registry
        The installed files are then moved to the local package directory

        :param force: Force the installation even if the dependency is already installed
        :param logger: The logger to use for logging installation output
        """
        logger = logger or _logger
        manifest = self.manifests.get(self.dependency.package_name)
        if manifest and manifest.version == self.dependency.version and not force:
            logger.info(
                f"Dependency {self.dependency.name} ({self.dependency.version}) is already installed"
            )
            return manifest

        for task in self.get_tasks():
            logger.info(f"{task}")
            out = task(device=self.handler)
            logger.info(out.decode())

        manifest = self.create_manifest()
        self.manifests.save(manifest)

        for item in manifest.items:
            task = MoveTask(
                source_file=self.artifact_root / item.path,
                target_file=self.install_root / item.path,
                workspace=self.workspace,
            )
            logger.info(f"{task}")
            task()

        return manifest

    def __call__(self, force=False, logger=None):
        logger = logger or _logger
        self.install(force=force, logger=logger)
        source = self.source
        deployment = source.create_deployment()
        if deployment:
            deployment(logger=logger)


@dataclass(frozen=True, eq=True)
class ToolInstaller(DependencyInstaller):
    """
    Models a tool installation process
    Tools are only installed to the tools directory and are not uploaded to the device
    """
    @property
    def artifact_root(self):
        return self.workspace.get_artifact_path("tools") / self.dependency.package_name

    @property
    def install_root(self):
        return self.workspace.get_tool_path()

    @property
    def manifests(self):
        return ToolManifestRegistry(workspace=self.workspace, asset_type="source")

    @property
    def source(self) -> WorkspaceSource:
        return WorkspaceToolSource(workspace=self.workspace, source=self.dependency)



"""
Filesystem operations to be run on the remote device.
Compatible micropython code
"""
import os


def resolve_path(relative_path):
    """
    Resolve a relative path to an absolute path.
    """
    if relative_path.startswith("/"):
        return relative_path
    # Get the current working directory
    cwd = os.getcwd()
    # Split the working directory and the relative path into components
    path_parts = cwd.split("/") + relative_path.split("/")

    # Resolve '.' and '..' manually
    resolved_parts = []
    for part in path_parts:
        if part == "." or part == "":
            continue
        elif part == "..":
            if resolved_parts:
                resolved_parts.pop()
        else:
            resolved_parts.append(part)

    # Join the resolved parts to form the absolute path
    absolute_path = "/" + "/".join(resolved_parts)
    return absolute_path


def iwalk(parent):
    """
    Recursively walk the filesystem starting from the parent directory.
    Yields files starting from the innermost directory.
    Folders are yielded after the files they contain to allow for easier deletion.
    """
    stack = [parent]
    visited = set()
    while stack:
        current = stack[-1]
        if current in visited:
            yield current
            stack.pop()
            continue

        try:
            os.stat(current)
        except OSError:
            stack.pop()
            continue

        for entry in os.ilistdir(current):
            name, etype = entry[0], entry[1]
            is_dir = etype == 0x4000

            path = f"{current.rstrip('/')}/{name}"
            if is_dir:
                stack.append(path)
            else:
                yield path

        visited.add(current)


def walk(parent):
    return list(iwalk(parent))


def rmtree(parent):
    """
    Recursively delete the directory and all its contents.
    """
    files = []
    cwd = os.getcwd()
    for path in iwalk(parent):
        path = resolve_path(path)
        if path.startswith(cwd) or cwd.startswith(path):
            continue
        os.remove(path)
        files.append(path)

    return files



"""
Core plugin provider for Caiman.
Provides the core plugins for the build tool.
"""
from caiman.plugins.base import PluginProvider
from caiman.plugins.workspace import WorkspacePlugin


class CorePluginProvider(PluginProvider):
    def get_plugins(self, config):
        from caiman.plugins.builder import ApplicationBuilderPlugin
        from caiman.plugins.deploy import DeployPlugin
        from caiman.plugins.fs import FileSystemPlugin
        from caiman.plugins.installer import MIPInstallerPlugin
        from caiman.plugins.runner import RunnerPlugin

        return [
            WorkspacePlugin(config=config),
            ApplicationBuilderPlugin(config=config),
            DeployPlugin(config=config),
            FileSystemPlugin(config=config),
            RunnerPlugin(config=config),
            MIPInstallerPlugin(config=config),
        ]



import subprocess
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

from caiman.config import Workspace
from caiman.proc.base import MicroPythonProcess


@dataclass(frozen=True, eq=True)
class Task(ABC):
    """
    Class for defining a copy task between two paths.
    """

    workspace: Workspace

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


@dataclass(frozen=True, eq=True)
class CopyTask(Task):
    source_file: Path
    target_file: Path

    @property
    def rel_source_path(self) -> Path:
        """
        The relative source path."""
        return self.workspace.get_relative_path(self.source_file)

    @property
    def rel_target_path(self) -> Path:
        """
        The relative target path."""
        return self.workspace.get_relative_path(self.target_file)

    def __call__(self, *args, **kwargs):
        self.target_file.parent.mkdir(parents=True, exist_ok=True)
        self.target_file.write_bytes(self.source_file.read_bytes())
        return self.target_file

    def __str__(self):
        return f"[{self.__class__.__name__}] {self.rel_source_path} -> {self.rel_target_path}"


@dataclass(frozen=True, eq=True)
class MoveTask(CopyTask):
    def __call__(self, *args, **kwargs):
        self.target_file.parent.mkdir(parents=True, exist_ok=True)
        self.source_file.rename(self.target_file)
        return self.target_file


@dataclass(frozen=True, eq=True)
class CompileTask(CopyTask):
    """
    Class for defining a compile task between two paths.
    """

    def __call__(self, *args, **kwargs):
        self.target_file.parent.mkdir(parents=True, exist_ok=True)
        command = [
            sys.executable,
            "-m",
            "mpy_cross_v6",
            str(self.source_file),
            "-o",
            str(self.target_file),
        ]
        subprocess.run(command, check=True, stderr=subprocess.PIPE)
        return self.target_file


@dataclass(frozen=True, eq=True)
class MIPTask(Task):
    """
    Class for defining a MIP task between two paths.
    """

    index: str
    packages: tuple
    root: Path
    target: Path = ""

    def __call__(self, *args, device: MicroPythonProcess, **kwargs):
        local_parent = self.root / self.target
        local_parent.mkdir(parents=True, exist_ok=True)
        target = self.root / self.target
        return device.mip_install(
            index=self.index,
            target=str(target),
            packages=dict(self.packages),
            no_mpy=True,
        )

    @classmethod
    def from_package_dict(
        cls, packages: dict, workspace: Workspace, index: str, root: Path, target: Path
    ):
        return [
            MIPTask(
                workspace=workspace,
                index=index,
                packages=tuple(packages.items()),
                root=root,
                target=target,
            )
        ]

    def __str__(self):
        local_target = self.root / self.target
        package_info = ", ".join(map("@".join, self.packages))
        return f"[{self.__class__.__name__}] {package_info} -> {local_target}"



import dataclasses
import logging

from caiman.config import DEFAULT_CONF_FILE, Command, Config, get_project_init_fields
from caiman.plugins.base import Goal, Plugin, fail

_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class WorkspaceCommand:
    pass


class WorkspaceInitGoal(Goal):
    def __init__(self, config: Config):
        self.config = config

    @property
    def help(self):
        return "Initialize a new workspace"

    @property
    def name(self):
        return "init"

    def get_schema(self):
        return WorkspaceCommand

    def __call__(self, command: Command):
        if DEFAULT_CONF_FILE.exists() and not command.force:
            fail("Config file already exists")
        else:
            _logger.info(f"Updating config file: {DEFAULT_CONF_FILE}")

        _logger.info("Project details:")
        print(f"Root: {self.config.workspace.root}")
        app = _updated_config_from_input(self.config.application)

        _logger.info("Workspace structure:")
        workspace = _updated_config_from_input(self.config.workspace)
        self.config = dataclasses.replace(
            self.config, application=app, workspace=workspace
        )
        self.config.save(path=DEFAULT_CONF_FILE)
        build_path = self.config.workspace.get_build_path()
        if not build_path.exists():
            build_path.mkdir(parents=True)
            _logger.info(f"Build directory created: {build_path}")

        package_path = self.config.workspace.get_package_path()
        if not package_path.exists():
            package_path.mkdir(parents=True)
            _logger.info(f"Local package directory created: {package_path}")

        tool_path = self.config.workspace.get_tool_path()
        if not tool_path.exists():
            tool_path.mkdir(parents=True)
            _logger.info(f"Local tools directory created: {tool_path}")

        for source in self.config.sources:
            source_path = self.config.workspace.get_path(source.parent)
            if not source_path.exists():
                source_path.mkdir(parents=True)
                _logger.info(f"Source directory created: {source_path}")

        _logger.info("Project initialized")


class WorkspacePlugin(Plugin):
    """
    Plugin to handle configuration files
    """

    def get_goals(self):
        return (WorkspaceInitGoal(config=self.config),)


def _updated_config_from_input(config: Config):
    init_fields = get_project_init_fields(config)
    update_dict = {}
    for init_field in init_fields:
        current_value = getattr(config, init_field.name)
        new_value = (
            input(f"{init_field.metadata['label']} [{current_value}]:") or current_value
        )
        update_dict[init_field.name] = new_value

    return dataclasses.replace(config, **update_dict)



"""
Plugins for installing dependencies and tools using package managers
"""
from dataclasses import dataclass

from caiman.config import Command, Dependency
from caiman.installer import DependencyInstaller, ToolInstaller
from caiman.plugins.base import Goal, Plugin, fail, param


@dataclass
class InstallCommand:
    dependency: str = param("Name of the dependency: <package>@<version>")
    scope: str = param(
        "Scope of the dependency: {dependencies,tools}", default="dependencies"
    )
    channel: str = param("Channel to install the dependency from", default="")
    reinstall: bool = param("Reinstall the dependency", default=False)

    @property
    def package(self):
        return self.dependency.split("@")[0]

    @property
    def version(self):
        return self.dependency.split("@")[1] if "@" in self.dependency else None


class InstallGoal(Goal):
    @property
    def help(self):
        return "Install a dependency or tool in the local workspace"

    @property
    def name(self):
        return "install"

    def get_schema(self):
        return InstallCommand

    def __call__(self, command: Command):
        command = InstallCommand(**command.params)
        if not command.version:
            fail("Dependency version is required. Use the format <package>@<version>")

        kwargs = dict(name=command.package, version=command.version)
        if command.channel:
            kwargs["channel"] = command.channel

        dep = Dependency(**kwargs)
        if command.scope == "dependencies":
            installer = DependencyInstaller(dependency=dep, config=self.config)
        elif command.scope == "tools":
            installer = ToolInstaller(dependency=dep, config=self.config)
        else:
            raise ValueError(f"Invalid scope: {command.scope}")
        installer(force=command.reinstall, logger=self._logger)


class MIPInstallerPlugin(Plugin):
    def get_goals(self):
        return [InstallGoal(config=self.config)]



"""
Plugin loader for caiman
"""

from importlib import import_module
from typing import List

from caiman.config import Config
from caiman.plugins.base import Plugin, PluginProvider
from caiman.plugins.core import CorePluginProvider
from caiman.plugins.workspace import WorkspacePlugin


def get_pre_init_plugins(config) -> List[Plugin]:
    """
    Get plugins that should run before the config is loaded
    """
    return [WorkspacePlugin(config)]


def load_plugins(config: Config) -> List[Plugin]:
    """
    Load plugins from a list of class references
    """
    plugins = []
    class_refs = config.workspace.plugins
    if not class_refs:
        class_refs = [f"{CorePluginProvider.__module__}.{CorePluginProvider.__name__}"]

    for class_ref in class_refs:
        module_name, class_name = class_ref.rsplit(".", 1)
        module = import_module(module_name)
        class_obj = getattr(module, class_name)
        if issubclass(class_obj, PluginProvider):
            plugins.extend(class_obj().get_plugins(config))
        elif issubclass(class_obj, Plugin):
            plugin = class_obj(config=config)
            plugins.append(plugin)
        else:
            raise TypeError(
                f"{class_obj} is not a subclass of Plugin or PluginProvider"
            )

    return plugins



"""
Data models for Caiman configuration artifacts.
"""
import os
from dataclasses import asdict, dataclass, field, fields
from functools import lru_cache
from pathlib import Path
from typing import List

import dacite
import pathspec
import yaml

IGNORES = [
    ".git",
    ".vscode",
    "**/__pycache__",
    "**/*.pyc",
]


DEFAULT_CONF_FILE = Path.cwd() / "caiman.yaml"


def config_field(
    default=IGNORES,
    default_factory=IGNORES,
    label=None,
    project_init=False,
    metadata=None,
):
    """
    A dataclass field with metadata for configuration that can be parsed by the
    project initialization command.
    """
    metadata = metadata or {}
    metadata["label"] = label
    metadata["project_init"] = project_init
    if default is not IGNORES:
        return field(default=default, metadata=metadata)
    if default_factory is not IGNORES:
        return field(default_factory=default_factory, metadata=metadata)


def get_field_label(field):
    """
    Get the label for the given field from the project initialization goal
    """
    return field.metadata.get("label", field.name)


def get_project_init_fields(cls):
    return [f for f in fields(cls) if f.metadata.get("project_init")]


@dataclass
class Command:
    """
    Models a high level command from the command line interface.
    """
    goal: str
    params: dict = field(default_factory=dict)
    force: bool = False


@dataclass
class Application:
    """
    Models the project application metadata.
    """
    name: str = config_field("", label="Project name", project_init=True)
    version: str = config_field("0.0.1", label="Project version", project_init=True)
    author: str = config_field("", label="Author", project_init=True)


@dataclass
class Device:
    """
    Configuration for working with a device.
    """
    port: str = ""


@lru_cache
def get_ignore_patterns(root: str) -> pathspec.PathSpec:
    """
    Get the ignore patterns for the .gitignore file in the workspace root.
    """
    ignore_file = Path(root) / ".gitignore"
    if ignore_file.exists():
        lines = ignore_file.read_text().splitlines()
        return pathspec.gitignore.GitIgnoreSpec.from_lines(lines)


class ConfigElement:
    """
    Base class for configuration elements.
    """
    def validate(self):
        pass


@dataclass(frozen=True, eq=True)
class Workspace(ConfigElement):
    """
    Configuration for the project workspace
    Includes paths for build, packages, tools, and plugins as a list of additional
    user-defined plugins for the build system
    """
    root: str
    build: str = config_field("build/board", label="Build directory", project_init=True)
    packages: str = config_field(
        "venv/mip-packages", label="Local MIP package directory", project_init=True
    )
    tools: str = config_field(
        "venv/tools", label="Local tools directory", project_init=True
    )
    plugins: List[str] = field(default_factory=list)
    extra_ignores: List[str] = field(default_factory=lambda: IGNORES)
    use_gitignore: bool = True

    def validate(self):
        """
        Validate the workspace configuration
        """
        for path in [
            self.get_path(),
            self.get_build_path(),
            self.get_package_path(),
            self.get_tool_path(),
        ]:
            if not path.is_relative_to(self.root):
                raise ValueError(f"Path {path} must be relative to workspace root")

    def get_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the workspace.
        """
        if Path(folder).is_absolute():
            raise ValueError(f"Folder path {folder} must not be absolute")
        return Path(self.root) / folder

    def get_build_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the build directory
        """
        return self.get_path(self.build) / folder

    def get_artifact_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the artifacts directory used by plugins
        to store temporary build artifacts.
        """
        return self.get_build_path("artifacts") / folder

    def get_manifest_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the manifests directory
        """
        return self.get_build_path("manifests") / folder

    def get_build_asset_path(self, is_frozen: bool, folder: str = "") -> Path:
        """
        Get the full path to a folder in the build directory for frozen or deployable assets.
        """
        return self.get_build_path("frozen" if is_frozen else "micropython") / folder

    def get_package_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the local packages directory. Dependencies
        are locally installed in this directory.
        """
        return self.get_path(self.packages) / folder

    def get_tool_path(self, folder: str = "") -> Path:
        """
        Get the full path to a folder in the local tools directory. Tools are locally
        installed in this directory.
        """
        return self.get_path(self.tools) / folder

    def get_ignore_patterns(self) -> pathspec.PathSpec:
        """
        Get the ignore patterns for the workspace by combining the .gitignore file
        patterns with the extra ignore patterns.
        """
        patterns = pathspec.gitignore.GitIgnoreSpec.from_lines(self.extra_ignores)
        root_patterns = get_ignore_patterns(self.root)
        if root_patterns:
            patterns = root_patterns + patterns
        return patterns

    def get_relative_path(self, path: Path) -> Path:
        """
        Get the relative path to a file or directory in the workspace.
        """
        if not path.is_relative_to(self.root):
            raise ValueError(f"Path {path} is not relative to workspace root")
        return path.relative_to(self.root)


@dataclass(frozen=True, eq=True)
class Channel(ConfigElement):
    """
    Configuration for a package channel.
    """
    name: str = "micropython"
    index: str = "https://micropython.org/pi/v2"


def default_channels():
    return [Channel()]


@dataclass(frozen=True, eq=True)
class Target(ConfigElement):
    """
    Base class for build targets.
    """
    name: str

    def to_dict(self):
        return asdict(self)

    @property
    def container(self):
        return None

    @property
    def is_frozen(self):
        return False


@dataclass(frozen=True, eq=True)
class PythonTarget(Target):
    """
    Base class for Python build targets.
    """
    frozen: bool = False
    compile: bool = True

    @property
    def is_frozen(self):
        return self.frozen


@dataclass(frozen=True, eq=True)
class FileSource(Target):
    """
    Base class for file sources.
    """
    files: List[str] = field(default_factory=list)
    parent: str = ""
    version: str = ""

    @property
    def package_name(self):
        return os.path.join(*self.name.split(":"))

    @property
    def container(self):
        return "micropython"

    def validate(self):
        if Path(self.parent).is_absolute():
            raise ValueError(f"Parent path {self.parent} must be relative")
        for file in self.files:
            if Path(file).is_absolute():
                raise ValueError(f"File path {file} must be relative")


def _default_python_path_patterns():
    return ["**/*.py"]


@dataclass(frozen=True, eq=True)
class PythonSource(FileSource, PythonTarget):
    """
    Configuration for a Python source.
    """
    files: List[str] = field(default_factory=_default_python_path_patterns)

    @classmethod
    def default_sources(cls):
        return [PythonSource(name="micropython", parent="micropython", compile=True)]

    @property
    def container(self):
        return super().container if not self.frozen else "frozen"


@dataclass(frozen=True, eq=True)
class Dependency(PythonSource):
    name: str
    version: str = "latest"
    channel: str = None
    files: List[str] = field(default_factory=list)


@dataclass
class Config:
    version: str = "0.0.1"
    device: Device = field(default_factory=Device)
    application: Application = field(default_factory=Application)
    workspace: Workspace = field(default_factory=Workspace)
    channels: List[Channel] = field(default_factory=default_channels)
    sources: List[PythonSource] = field(default_factory=PythonSource.default_sources)
    dependencies: List[Dependency] = field(default_factory=list)
    resources: List[FileSource] = field(default_factory=list)
    tools: List[Dependency] = field(default_factory=list)

    @classmethod
    def default(cls) -> "Config":
        return cls(workspace=Workspace(root=str(Path.cwd())))

    @classmethod
    def load(cls, path: str = "") -> "Config":
        path = path or DEFAULT_CONF_FILE
        cfg = yaml.safe_load(open(path)) or {}
        cfg.setdefault("workspace", {})["root"] = str(Path(path).resolve().parent)
        config = dacite.from_dict(data_class=cls, data=cfg)
        config.validate()
        return config

    def validate(self):
        self.workspace.validate()
        for source in self.sources:
            source.validate()
        for dep in self.dependencies:
            dep.validate()
        for res in self.resources:
            res.validate()
        for tool in self.tools:
            tool.validate()

    def save(self, path: str = "") -> None:
        self.validate()
        path = path or DEFAULT_CONF_FILE
        with open(path, "w") as f:
            d = asdict(self)
            d.get("workspace", {}).pop("root", None)
            yaml.dump(d, f, sort_keys=False)

    @property
    def root_path(self):
        return self.workspace.root

    def get_channel(self, name=None) -> Channel:
        if name:
            channel = next((c for c in self.channels if c.name == name), None)
            if not channel:
                raise ValueError(f"Channel {name} not found")
        else:
            channel = self.channels[0]
        return channel



import importlib
import json
import logging
import subprocess
import sys
from pathlib import Path
from typing import Callable

from caiman.config import Device
from caiman.proc.base import CommandError, MicroPythonProcess, follow_subprocess

_logger = logging.getLogger("device")


class DeviceMicroPythonProcess(MicroPythonProcess):
    def __init__(self, config: Device, mount_path: str = None):
        self.config = config
        self.mount_path = mount_path

    def mip_install(self, index: str, target: str, packages: dict, no_mpy):
        command = ["mip"]
        command += ["--no-mpy"] if no_mpy else []
        command += ["--index", self.index, "--target", str(target), "install"] + [
            "@".join(package) for package in self.packages
        ]
        return self.run_mp_remote_cmd(*command)

    def run_mp_remote_cmd(self, *args, cwd=None, follow=False):
        # Start a subprocess of the same Python interpreter
        cmd = [sys.executable, "-m", "mpremote"]
        if self.config.port:
            cmd.extend(["connect", self.config.port])

        if self.mount_path:
            cmd.extend(["mount", "-l", str(self.mount_path), "+"])

        cmd.extend(list(args))
        cmd.extend(["+", "disconnect"])
        _logger.debug(f"Running command: {' '.join(cmd)}")
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd
        )
        if follow:
            follow_subprocess(proc)
            sys.exit(proc.returncode)

        out, err = proc.communicate()
        if proc.returncode != 0:
            raise CommandError(cmd[-1], stdout=out, stderr=err)
        return out

    def run_code(self, code, follow=False):
        if isinstance(code, list):
            code = ";".join(code)
        args = ["soft-reset"]
        if self.mount_path:
            args.extend(["mount", "-l", str(self.mount_path), "+"])

        args.extend(["exec", code])
        return self.run_mp_remote_cmd(*args, follow=follow)

    def run_vfs_python_func(self, func: Callable, **kwargs):
        kwarg_json = json.dumps(kwargs)
        kwarg_decode = f"json.loads('{kwarg_json}')"
        func_module = importlib.import_module(func.__module__)
        if not func_module or not func_module.__file__:
            raise ImportError(f"Could not import module {func.__module__}")

        import_mod = func.__module__.split(".", 2)[-1]
        mount_path = Path(func_module.__file__).parent.parent
        func_name = func.__name__

        code = []
        code.append(f"from {import_mod} import {func_name}")
        code.append("import json")
        code.append(f"print(':::' + json.dumps({func_name}(**{kwarg_decode})))")

        output = (
            DeviceMicroPythonProcess(config=self.config, mount_path=str(mount_path))
            .run_code(code)
            .decode()
        )
        lines = output.splitlines()
        lines = [line[3:] for line in lines if line.startswith(":::")]
        output = lines[-1]
        result = json.loads(output)
        return result



"""
The manifest module provides classes for collecting and tracking manifests of files that
compose a build target.
Manifests are generated for resources, sources, dependencies, and tools.
"""
import json
from dataclasses import asdict, dataclass, field
from hashlib import sha1
from pathlib import Path
from typing import List

import dacite

from caiman.config import Workspace


@dataclass(frozen=True, eq=True)
class ManifestItem:
    """
    Class for defining a file in a manifest. All paths are relative as manifests can be interpreted
    in different contexts (e.g: temporary artifact locations, local package directories, etc.)
    """
    path: str
    sha1: str
    size: int

    @classmethod
    def create(cls, relative_path: Path, source_root: Path):
        """
        Create a manifest item from a relative path and source root.
        :param relative_path: The relative path of the file.
        :param source_root: The root directory of the source to which the path is relative.
        """
        return cls(
            path=str(relative_path),
            sha1=str(sha1(Path(source_root / relative_path).read_bytes()).hexdigest()),
            size=Path(source_root / relative_path).stat().st_size,
        )

    @classmethod
    def from_paths(cls, paths: List[Path], source_root: Path):
        """
        Create a list of manifest items from a list of relative paths and source root.
        :param paths: The list of relative paths of the files.
        :param source_root: The root directory of the source to which the paths are relative.
        """
        return [cls.create(path, source_root) for path in paths]

    def is_file_changed(self, path: Path) -> bool:
        """
        Check if the file tracked by this manifest item has changed with respect to the file at the given path.
        """
        return (
            not path.exists()
            or self.size != path.stat().st_size
            or self.sha1 != sha1(path.read_bytes()).hexdigest()
        )


@dataclass(frozen=True, eq=True)
class Manifest:
    """
    Class for defining a manifest of files that compose a build target.
    """
    name: str
    version: str = ""
    items: List[ManifestItem] = field(default_factory=list)

    def __iter__(self):
        return (Path(item.path) for item in self.items)


@dataclass(frozen=True, eq=True)
class ManifestRegistry:
    """
    Class for defining a registry of manifests for a particular asset type.
    Asset types include source code, and target files
    """
    workspace: Workspace
    asset_type: str = "source"

    @property
    def folder(self) -> str:
        """
        Parent folder for the manifest files. This is used to organize the manifest files for different
        types of build targets.
        """
        return ""

    def get_manifest_path(self, package: str) -> Path:
        """
        Get the path to the manifest file for the source files.
        :param package: The name of the package for which the manifest is being created.
        """
        base_name = f"{package}-{self.asset_type}.json"
        return self.workspace.get_manifest_path(self.folder) / base_name

    def get(self, package: str) -> Manifest:
        """
        Load the manifest for the source files.
        :param package: The name of the package for which the manifest is being created.
        """
        path = self.get_manifest_path(package)
        if path.exists():
            json_manifest = json.loads(path.read_text()).get(package, {})
            return dacite.from_dict(
                data_class=Manifest,
                data=dict(parent=self.folder, name=package, **json_manifest),
            )

    def save(self, manifest: Manifest):
        """
        Save the manifest for the source files.
        """
        manifest_dict = asdict(manifest)
        json_dict = {
            manifest_dict.pop("name"): manifest_dict,
        }
        path = self.get_manifest_path(manifest.name)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(json_dict, indent=2))
        return manifest


@dataclass(frozen=True, eq=True)
class DependencyManifestRegistry(ManifestRegistry):
    """
    Class for defining a registry of manifests for the installed dependencies.
    """
    @property
    def folder(self) -> str:
        return "dependencies"


@dataclass(frozen=True, eq=True)
class SourceManifestRegistry(ManifestRegistry):
    """
    Class for defining a registry of manifests for the source files.
    """
    @property
    def folder(self) -> str:
        return "sources"


@dataclass(frozen=True, eq=True)
class ResourceManifestRegistry(ManifestRegistry):
    """
    Class for defining a registry of manifests for the resource files.
    """
    @property
    def folder(self) -> str:
        return "resources"


@dataclass(frozen=True, eq=True)
class ToolManifestRegistry(ManifestRegistry):
    """
    Class for defining a registry of manifests for the tools.
    """
    @property
    def folder(self) -> str:
        return "tools"



"""
Base classes for defining plugins and goals.
"""
import dataclasses
import logging
import sys
from abc import ABC, abstractmethod
from typing import Tuple, Type

from caiman.config import Command, Config


class Goal(ABC):
    """
    A goal is a specific task that can be executed by a plugin.
    """
    def __init__(self, config: Config):
        """
        Initialize the goal with the given project configuration.
        :param config: The project configuration
        """
        self.config = config
        self._logger = logging.getLogger(self.name)

    @property
    @abstractmethod
    def help(self):
        """
        Return the help message for the goal.
        """

    @property
    @abstractmethod
    def name(self):
        """
        Return the name of the goal for the command line interface.
        """
        return self.__class__.__name__

    @abstractmethod
    def get_schema(self) -> Type:
        """
        Return the schema for the command parameters.
        """

    @abstractmethod
    def __call__(self, command: Command):
        """
        Execute the goal with the given command.
        """

    def fail(self, message):
        """
        Exit the program with an error message.
        """
        fail(f"[{self.name}] {message}")

    def info(self, message):
        """
        Log an informational message
        """
        self._logger.info(f"{message}")


class Plugin(ABC):
    """
    A plugin is a collection of goals that can be executed by the build tool.
    """
    def __init__(self, config: Config):
        """
        Initialize the plugin with the given project configuration.
        :param config: The project configuration
        """
        self.config = config

    @abstractmethod
    def get_goals(self) -> Tuple[Goal]:
        """
        Return the goals provided by the plugin.
        """

    @property
    def name(self):
        """
        Return the name of the plugin.
        """
        return self.__class__.__name__


class PluginProvider(ABC):
    """
    A plugin provider is a factory for creating plugins.
    """
    @abstractmethod
    def get_plugins(self, config: Config):
        return ()


def fail(message):
    """
    Exit the program with an error message.
    """
    print(message)
    sys.exit(1)


def param(
    help, default=dataclasses.MISSING, default_factory=dataclasses.MISSING, **kwargs
):
    """
    A dataclass field with help text indexed by the command line parser.
    """
    return dataclasses.field(
        metadata={"help": help},
        default=default,
        default_factory=default_factory,
        **kwargs
    )



"""
Classes to model sources of build targets in a workspace.
"""

from dataclasses import dataclass
from pathlib import Path

from pathspec import PathSpec

from caiman.config import Dependency, FileSource, PythonSource, Workspace
from caiman.deployment import (
    DependencyDeployment,
    Deployment,
    PythonDeployment,
)
from caiman.manifest import (
    DependencyManifestRegistry,
    Manifest,
    ManifestItem,
    ResourceManifestRegistry,
    SourceManifestRegistry,
    ToolManifestRegistry,
)


@dataclass(frozen=True, eq=True)
class WorkspaceSource:
    """
    Base class for defining a source of files in a workspace.
    """

    workspace: Workspace
    source: FileSource

    def __post_init__(self):
        if Path(self.source.parent).is_absolute():
            raise ValueError(
                f"Source parent directory path must be relative: {self.source}"
            )

    @property
    def name(self):
        """
        The name of the source files."""
        return self.source.name

    @property
    def root(self) -> Path:
        """
        The root directory of the source files."""
        return self.workspace.get_path(self.source.parent)

    @property
    def ignores(self) -> PathSpec:
        """
        The ignore patterns for the source files."""
        return self.workspace.get_ignore_patterns()

    @property
    def manifests(self):
        """
        The registry of manifests for the source files.
        """
        return ResourceManifestRegistry(workspace=self.workspace, asset_type="source")

    def create_manifest(self):
        """
        Create a manifest for the source files.
        """
        return Manifest(
            name=self.source.package_name,
            version=self.source.version,
            items=ManifestItem.from_paths(list(self), self.root),
        )

    def get_manifest(self):
        """
        Get the manifest for the source files.
        """
        return self.manifests.get(self.source.package_name)

    def create_deployment(self):
        """
        Create a deployment for the source files.
        """
        return Deployment(
            source_path=self.root,
            source_manifest=self.get_manifest(),
            workspace=self.workspace,
            is_frozen=self.source.is_frozen,
            compile=False,
        )

    def __iter__(self):
        """
        Generator for the source files."""
        patterns = self.source.files if self.source.files else ["**/*"]
        for pattern in patterns:
            for path in Path(self.root).rglob(pattern):
                if path.is_file():
                    if not self.ignores or not self.ignores.match_file(str(path)):
                        yield path.relative_to(self.root)


@dataclass(frozen=True, eq=True)
class WorkspacePythonSource(WorkspaceSource):
    source: PythonSource

    @property
    def manifests(self):
        return SourceManifestRegistry(workspace=self.workspace, asset_type="source")

    def create_deployment(self):
        """
        Create a deployment for the source files.
        """
        return PythonDeployment(
            source_path=self.root,
            source_manifest=self.get_manifest(),
            workspace=self.workspace,
            is_frozen=self.source.is_frozen,
            compile=self.source.compile,
        )


@dataclass(frozen=True, eq=True)
class WorkspaceDependencySource(WorkspaceSource):
    source: Dependency

    @property
    def ignores(self) -> PathSpec:
        return None

    @property
    def manifests(self):
        return DependencyManifestRegistry(workspace=self.workspace, asset_type="source")

    def __iter__(self):
        return iter(self.get_manifest() or [])

    @property
    def root(self):
        return self.workspace.get_package_path()

    def create_deployment(self):
        """
        Create a deployment for the source files.
        """
        return DependencyDeployment(
            source_path=self.workspace.get_package_path(),
            source_manifest=self.get_manifest(),
            workspace=self.workspace,
            is_frozen=self.source.is_frozen,
            compile=self.source.compile,
        )


@dataclass(frozen=True, eq=True)
class WorkspaceToolSource(WorkspaceDependencySource):
    @property
    def manifests(self):
        return ToolManifestRegistry(workspace=self.workspace, asset_type="source")

    @property
    def root(self) -> Path:
        return self.workspace.get_tool_path()

    def create_deployment(self):
        """
        Create a deployment for the source files.
        """
        return None



import subprocess
import sys

from caiman.proc.base import (
    CommandError,
    MicroPythonProcess,
    find_micropython_path,
    follow_subprocess,
)


class LocalMicroPythonProcess(MicroPythonProcess):
    def __init__(self, executable=None):
        self.executable = executable or find_micropython_path()

    def mip_install(self, index: str, target: str, packages: dict, no_mpy):
        code = ["import mip"]
        for name, version in packages.items():
            code.append(
                f"mip.install('{name}', version='{version}', index='{index}', target='{target}',mpy={not no_mpy})"
            )
        return self.run_code(code, follow=False)

    def run_code(self, code, cwd=None, follow=False):
        # Run the given lines of MicroPython code
        cmd = [self.executable, "-c", ";".join(code)]
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd
        )
        if follow:
            follow_subprocess(proc)
            sys.exit(proc.returncode)

        out, err = proc.communicate()
        if proc.returncode != 0:
            raise CommandError(cmd[-1], stdout=out, stderr=err)

        return out



"""
Models deployments of source files and locally installed dependencies to the build directory
for future deployment to a target device.
"""
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator

from caiman.config import Workspace
from caiman.manifest import (
    DependencyManifestRegistry,
    Manifest,
    ManifestItem,
    ResourceManifestRegistry,
    SourceManifestRegistry,
    ToolManifestRegistry,
)
from caiman.task import CompileTask, CopyTask, Task

_logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True)
class Deployment:
    """
    Models a deployment of source files and locally installed dependencies to the build directory
    """
    source_path: Path
    source_manifest: Manifest
    workspace: Workspace
    is_frozen: bool
    compile: bool

    @property
    def manifests(self):
        """
        The manifest registry for the deployed files
        """
        return ResourceManifestRegistry(
            workspace=self.workspace,
            asset_type="target",
        )

    @property
    def path(self) -> Path:
        """
        The root directory of the target files."""
        return self.workspace.get_build_asset_path(is_frozen=self.is_frozen)

    def get_tasks(self) -> Iterator[Task]:
        """
        Generator for the source and target file paths."""
        for source_item in self.source_manifest.items:
            source_path = self.source_path / source_item.path

            target_path = self.path / Path(source_item.path)
            if target_path.suffix == ".py" and self.compile:
                target_path = target_path.with_suffix(".mpy")
                yield CompileTask(
                    source_file=source_path,
                    target_file=target_path,
                    workspace=self.workspace,
                )
            else:
                yield CopyTask(
                    source_file=source_path,
                    target_file=target_path,
                    workspace=self.workspace,
                )

    def __iter__(self) -> Iterator[Path]:
        """
        The manifest items for the target files."""
        for task in self.get_tasks():
            yield task.target_file.relative_to(self.path)

    def __call__(self, logger=None):
        logger = logger or _logger
        items = []
        for task in self.get_tasks():
            logger.info(f"{task}")
            target_file = task()
            rel_path = target_file.relative_to(self.path)
            items.append(ManifestItem.create(rel_path, self.path))

        return Manifest(
            name=self.source_manifest.name,
            version=self.source_manifest.version,
            items=items,
        )


@dataclass(frozen=True, eq=True)
class PythonDeployment(Deployment):
    """
    Models a deployment of Python source files to the build directory
    """
    @property
    def manifests(self):
        return SourceManifestRegistry(
            workspace=self.workspace,
            asset_type="target",
        )


@dataclass(frozen=True, eq=True)
class DependencyDeployment(Deployment):
    """
    Models a deployment of locally installed dependencies to the build directory
    """
    @property
    def manifests(self):
        return DependencyManifestRegistry(
            workspace=self.workspace,
            asset_type="target",
        )


@dataclass(frozen=True, eq=True)
class ToolDeployment(Deployment):
    """
    Models a deployment of tool files to the tools directory
    """
    @property
    def manifests(self):
        return ToolManifestRegistry(
            workspace=self.workspace,
            asset_type="target",
        )



"""
Plugins that upload build artifacts to the target device
"""
from dataclasses import dataclass

from caiman.config import Command, Config
from caiman.device import FileSystem
from caiman.plugins.base import Goal, Plugin
from caiman.proc.device import DeviceMicroPythonProcess


@dataclass
class DeployCommand:
    pass


class DeployGoal(Goal):
    def __init__(self, config: Config, fs: FileSystem):
        super().__init__(config)
        self._fs = fs

    @property
    def help(self):
        return "Deploy build artifacts to the target device"

    @property
    def name(self):
        return "deploy"

    def get_schema(self):
        return DeployCommand

    def __call__(self, command: Command):
        mp_root_path = self.config.workspace.get_build_asset_path(is_frozen=False)
        paths = mp_root_path.glob("*")
        for path in paths:
            rel_path = str(path.relative_to(mp_root_path))
            self.info(f"Uploading {rel_path}")
            self._fs.upload(src=rel_path, dst="", cwd=mp_root_path)


class DeployPlugin(Plugin):
    def __init__(self, config: Config):
        super().__init__(config=config)
        self._fs = FileSystem(device=DeviceMicroPythonProcess(config=config.device))

    def get_goals(self):
        return (DeployGoal(config=self.config, fs=self._fs),)



"""
APIs for interacting with the device from the host
"""
import json
from pathlib import Path

from caiman.proc.base import CommandError
from caiman.proc.device import DeviceMicroPythonProcess
from caiman.remote.caiman import fs


class FileSystem:
    """
    Manage file system operations on the device.
    """

    def __init__(self, device: DeviceMicroPythonProcess):
        """
        :param device: DeviceHandler instance to run device commands
        """
        self.device = device

    def walk(self, path) -> list[str]:
        """
        Recursively walk the file system starting at the given path.
        :param path: Path to the parent directory to walk
        :return: List of file paths
        """
        return self.device.run_vfs_python_func(fs.walk, parent=path)

    def rmtree(self, path):
        """
        Recursively remove the given path.
        :param path: Path to remove
        """
        return self.device.run_vfs_python_func(fs.rmtree, parent=path)

    def upload(self, src, dst, cwd=None):
        """
        Upload a file or directory to the device.
        """
        cmd = ["fs", "cp"]
        full_src = Path(src) if cwd is None else Path(cwd) / src
        if full_src.is_dir():
            cmd.append("-r")

        dst = dst.strip("/")
        dst = f":{dst}"
        cmd.extend([src, dst])
        return self.device.run_mp_remote_cmd(*cmd, cwd=cwd)

    def mkdir(self, path):
        """
        Create the given path on the device.
        """
        parts = path.split("/")
        for i in range(1, len(parts)):
            subpath = "/".join(parts[: i + 1])
            try:
                self.device.run_mp_remote_cmd("mkdir", subpath)
                print(f"Created {subpath}")
            except CommandError:
                pass

    def get_file_contents(self, file_path):
        """
        Get the contents of a file on the device.
        """
        return self.device.run_mp_remote_cmd("cat", file_path)

    def get_json(self, file_path, ignore_missing=False):
        """
        Get the contents of a file on the device as a JSON object
        """
        try:
            contents = self.get_file_contents(file_path)
        except CommandError as e:
            if ignore_missing:
                contents = "{}"
            else:
                raise e
        return json.loads(contents)



from dataclasses import dataclass

from caiman.config import Command
from caiman.plugins.base import Goal, Plugin, param
from caiman.proc.device import DeviceMicroPythonProcess


@dataclass
class RunnerCommand:
    target: str = param("Python module to run on the device")


class RunnerGoal(Goal):
    def __init__(self, device: DeviceMicroPythonProcess):
        self._device = device

    @property
    def help(self):
        return "Run a Python module on the device"

    @property
    def name(self):
        return "run"

    def get_schema(self):
        return RunnerCommand

    def __call__(self, command: Command):
        command = RunnerCommand(**command.params)
        return self._device.run_main(command.target)


class RunnerPlugin(Plugin):
    def __init__(self, config):
        super().__init__(config)
        self._device = DeviceMicroPythonProcess(config=config.device)

    def get_goals(self):
        return (RunnerGoal(device=self._device),)



"""
Caiman build system for MicroPython

Entry point for the caiman build system. This script is responsible for parsing the command line arguments and
loading the plugins that match the command. It then runs the plugins in order to build the MicroPython firmware.

(c) 2024 Andrei Dumitrache
"""
import argparse
import dataclasses
import json
import logging
import os
from pathlib import Path
from typing import Tuple

from caiman.config import DEFAULT_CONF_FILE, Command, Config
from caiman.loader import get_pre_init_plugins, load_plugins
from caiman.plugins.base import Goal, Plugin, fail
from caiman.proc.base import CommandError

_logger = logging.getLogger("caiman")


def get_arg_parser(goals: Tuple[Plugin]):
    parser = argparse.ArgumentParser(description="caiman - MicroPython build tool")
    goal_parsers = parser.add_subparsers(
        dest="goal",
        help="The goal to execute",
        required=True,
        title="Goals",
    )

    parser.add_argument(
        "--silent", help="Enable verbose output", action="store_true", default=False
    )
    parser.add_argument(
        "--force",
        help="Force the execution of the command",
        action="store_true",
        default=False,
    )

    for goal in goals:
        goal_parser = goal_parsers.add_parser(name=goal.name, help=goal.help)
        schema = goal.get_schema()
        for field in dataclasses.fields(schema):
            kwargs = {}
            if field.default is not dataclasses.MISSING:
                kwargs["default"] = field.default
                if type(field.default) is bool:
                    kwargs["action"] = "store_true"
            elif field.default_factory is not dataclasses.MISSING:
                kwargs["default_factory"] = field.default_factory
            else:
                kwargs["required"] = True
            goal_parser.add_argument(
                f"--{field.name}",
                help=field.metadata.get("help"),
                dest=f"params.{field.name}",
                **kwargs,
            )

    return parser


def get_goals(plugins: Tuple[Plugin]) -> Tuple[Goal]:
    plugin_by_goal = {}
    goals = []
    for plugin in plugins:
        for goal in plugin.get_goals():
            if goal.name in plugin_by_goal:
                fail(
                    f"Goal {goal.name} provided by multiple plugins: {plugin_by_goal[goal.name].name}, {plugin.name}"
                )
            plugin_by_goal[goal.name] = plugin
            goals.append(goal)

    return tuple(goals)


def main():
    plugins = []
    config_file = os.getenv("CAIMAN_CONFIG", DEFAULT_CONF_FILE)
    if not Path(config_file).exists():
        config = Config.default()
        plugins.extend(get_pre_init_plugins(config))
    else:
        config = Config.load(config_file)
        plugins.extend(load_plugins(config))

    goals = get_goals(plugins)

    parser = get_arg_parser(goals)
    args = parser.parse_args()
    logging.basicConfig(
        level=logging.ERROR if args.silent else logging.INFO,
        format="%(levelname)s [%(name)s] %(message)s",
    )

    if not args.silent:
        print(
            """

▗▄▄▖ ▗▄▖ ▗▄▄▄▖▗▖  ▗▖ ▗▄▖ ▗▖  ▗▖
▐▌   ▐▌ ▐▌  █  ▐▛▚▞▜▌▐▌ ▐▌▐▛▚▖▐▌
▐▌   ▐▛▀▜▌  █  ▐▌  ▐▌▐▛▀▜▌▐▌ ▝▜▌
▝▚▄▄▖▐▌ ▐▌▗▄█▄▖▐▌  ▐▌▐▌ ▐▌▐▌  ▐▌
================================
MicroPython build system
        """
        )

    _logger.info(f"Plugins: {', '.join([plugin.name for plugin in plugins])}")
    goal = next((g for g in goals if g.name == args.goal), None)
    if not goal:
        fail(f"Goal {args.goal} not provided by any plugin")

    goal_kwargs = {
        k.split(".", 1)[1]: v for k, v in vars(args).items() if k.startswith("params.")
    }
    command = Command(goal=goal.name, params=goal_kwargs, force=args.force)
    _logger.info(f"[{goal.name}] Running command: {command}")
    try:
        output = goal(command)
        if isinstance(output, bytes):
            print(output.decode("utf-8"))
        elif output:
            print(json.dumps(output, indent=2))
    except CommandError as e:
        fail(f"Board command error: {e}")

    if not DEFAULT_CONF_FILE.exists():
        fail(f"Config file {DEFAULT_CONF_FILE} does not exist")


if __name__ == "__main__":
    main()



import select
import shutil
import sys
from abc import ABC, abstractmethod


class CommandError(Exception):
    def __init__(self, command, stdout, stderr):
        self.command = command
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        output = [f"command:\n{self.command}"]
        if self.stdout:
            output.append(f"stdout:\n{self.stdout.decode()}")
        if self.stderr:
            output.append(f"stderr:\n{self.stderr.decode()}")
        return "\n".join(output)


class MicroPythonProcess(ABC):
    @abstractmethod
    def mip_install(self, index: str, target: str, packages: dict, no_mpy):
        pass

    def run_main(self, module_name):
        return self.run_code(f"import {module_name};", follow=True)

    @abstractmethod
    def run_code(self, code, follow=False):
        pass


def follow_subprocess(proc):
    while True:
        reads = [proc.stdout.fileno(), proc.stderr.fileno()]
        ret = select.select(reads, [], [])

        for fd in ret[0]:
            if fd == proc.stdout.fileno():
                output = proc.stdout.readline()
                if output:
                    output = output.decode() if isinstance(output, bytes) else output
                    sys.stdout.write(output)
                    sys.stdout.flush()
            if fd == proc.stderr.fileno():
                error = proc.stderr.readline()
                if error:
                    error = error.decode() if isinstance(error, bytes) else error
                    sys.stderr.write(error)
                    sys.stderr.flush()

        if proc.poll() is not None:
            break


def find_micropython_path():
    # Find the MicroPython binary on the system path
    micropython_path = shutil.which("micropython")
    if micropython_path:
        return micropython_path
    else:
        raise FileNotFoundError("MicroPython binary not found in system PATH")



from abc import ABC
from dataclasses import dataclass

from caiman.config import Command
from caiman.device import FileSystem
from caiman.plugins.base import Goal, Plugin, param
from caiman.proc.device import DeviceMicroPythonProcess


@dataclass
class FSTargetCommand:
    target: str = param("Target path on the remote device")


class FileSystemGoal(Goal, ABC):
    def __init__(self, fs: FileSystem):
        self.fs = fs

    def get_schema(self):
        return FSTargetCommand


class WalkGoal(FileSystemGoal):
    @property
    def name(self):
        return "walk"

    @property
    def help(self):
        return "Recursively walk the filesystem"

    def __call__(self, command: Command):
        return self.fs.walk(FSTargetCommand(**command.params).target)


class RMTreeGoal(FileSystemGoal):
    @property
    def name(self):
        return "rmtree"

    @property
    def help(self):
        return "Recursively remove a directory"

    def __call__(self, command: Command):
        return self.fs.rmtree(FSTargetCommand(**command.params).target)


class UploadGoal(FileSystemGoal):
    @property
    def name(self):
        return "upload"

    @property
    def help(self):
        return "Upload a file or directory"

    def __call__(self, command: Command):
        return self.fs.upload(FSTargetCommand(**command.params).target)


class GetFileContentsGoal(FileSystemGoal):
    @property
    def name(self):
        return "cat"

    @property
    def help(self):
        return "Get the contents of a file"

    def __call__(self, command: Command):
        return self.fs.get_file_contents(FSTargetCommand(**command.params).target)


class MKDirGoal(FileSystemGoal):
    @property
    def name(self):
        return "mkdir"

    @property
    def help(self):
        return "Create a directory"

    def __call__(self, command: Command):
        return self.fs.mkdir(FSTargetCommand(**command.params).target)


class GetJsonGoal(FileSystemGoal):
    @property
    def name(self):
        return "json.load"

    @property
    def help(self):
        return "Get the contents of a file as JSON"

    def __call__(self, command: Command):
        return self.fs.get_json(FSTargetCommand(**command.params).target)


class FileSystemPlugin(Plugin):
    def __init__(self, config):
        super().__init__(config)
        self._fs = FileSystem(device=DeviceMicroPythonProcess(config=config.device))

    def get_goals(self):
        return (
            WalkGoal(fs=self._fs),
            RMTreeGoal(fs=self._fs),
            UploadGoal(fs=self._fs),
            MKDirGoal(fs=self._fs),
            GetFileContentsGoal(fs=self._fs),
            GetJsonGoal(fs=self._fs),
        )
